"""Device capabilities."""
from __future__ import annotations

import logging
from enum import IntEnum

from ..events import EventHandler, LandroidEvent


class DeviceCapability(IntEnum):
    """Available device capabilities."""

    EDGE_CUT = 1
    ONE_TIME_SCHEDULE = 2
    PARTY_MODE = 4
    TORQUE = 8


CAPABILITY_TO_TEXT = {
    DeviceCapability.EDGE_CUT: "Edge Cut",
    DeviceCapability.ONE_TIME_SCHEDULE: "One-Time-Schedule",
    DeviceCapability.PARTY_MODE: "Party Mode",
    DeviceCapability.TORQUE: "Motor Torque",
}


class Capability(int):
    """Class for handling device capabilities."""

    def __init__(self) -> None:
        """Initialize an empty capability list."""
        super().__init__()
        self.__int__ = 0
        self.ready = False

    def add(self, capability: DeviceCapability) -> None:
        """Add capability to the list."""
        if capability & self.__int__ == 0:
            self.__int__ = self.__int__ | capability

    def check(self, capability: DeviceCapability) -> bool:
        """Check if device has capability."""
        if capability & self.__int__ == 0:
            return False
        else:
            return True



"""Defines actions for the devices."""
from __future__ import annotations

import json

from ..exceptions import NoOneTimeScheduleError, NoPartymodeError, OfflineError
from ..helpers import get_logger
from .capability import DeviceCapability
from .mqtt import Command

_LOGGER = get_logger("mqtt")


class Actions:
    """Class for actions."""

    def home(self) -> None:
        """Stop the current task and go home.
        If the knifes was turned on when this is called, it will return home with knifes still turned on.

        Raises:
            OfflineError: Raised if the device is offline.
        """
        if self.online:
            logger = _LOGGER.getChild("command")
            logger.debug("Sending HOME command to %s", self.name)
            self.mqtt.command(self.name, Command.HOME)
        else:
            raise OfflineError("The device is currently offline, no action was sent.")

    def zonetraining(self) -> None:
        """Start the zone training task.

        Raises:
            OfflineError: Raised if the device is offline.
        """
        if self.online:
            logger = _LOGGER.getChild("command")
            logger.debug("Sending ZONETRAINING command to %s", self.name)
            self.mqtt.command(self.name, Command.ZONETRAINING)
        else:
            raise OfflineError("The device is currently offline, no action was sent.")

    def lock(self, enabled: bool) -> None:
        """Set the device locked state.

        Args:
            enabled (bool): True will lock the device, False will unlock the device.

        Raises:
            OfflineError: Raised if the device is offline.
        """
        if self.online:
            if enabled:
                logger = _LOGGER.getChild("command")
                logger.debug("Sending LOCK command to %s", self.name)
                self.mqtt.command(self.name, Command.LOCK)
            else:
                logger = _LOGGER.getChild("command")
                logger.debug("Sending UNLOCK command to %s", self.name)
                self.mqtt.command(self.name, Command.UNLOCK)
        else:
            raise OfflineError("The device is currently offline, no action was sent.")

    def restart(self):
        """Reboot the device baseboard.

        Raises:
            OfflineError: Raised if the device is offline.
        """
        if self.online:
            logger = _LOGGER.getChild("command")
            logger.debug("Sending RESTART command to %s", self.name)
            self.mqtt.command(self.name, Command.RESTART)
        else:
            raise OfflineError("The device is currently offline, no action was sent.")

    def safehome(self):
        """Stop and go home with the blades off

        Raises:
            OfflineError: Raised if the device is offline.
        """
        if self.online:
            logger = _LOGGER.getChild("command")
            logger.debug("Sending SAFEHOME command to %s", self.name)
            self.mqtt.command(self.name, Command.SAFEHOME)
        else:
            raise OfflineError("The device is currently offline, no action was sent.")

    def raindelay(self, rain_delay: str) -> None:
        """Set new rain delay.

        Args:
            rain_delay (str | int): Rain delay in minutes.

        Raises:
            OfflineError: Raised if the device is offline.
        """
        if self.online:
            if not isinstance(rain_delay, str):
                rain_delay = str(rain_delay)
            msg = {"rd": rain_delay}
            self.mqtt.send(self.name, str(msg))
        else:
            raise OfflineError("The device is currently offline, no action was sent.")

    def toggle_schedule(self, enable: bool) -> None:
        """Turn on or off the schedule.

        Args:
            enable (bool): True is enabling the schedule, Fasle is disabling the schedule.

        Raises:
            OfflineError: Raised if the device is offline.
        """
        if self.online:
            if enable:
                msg = {"sc": {"m": 1}}
            else:
                msg = {"sc": {"m": 0}}

            self.mqtt.send(self.name, str(msg))
        else:
            raise OfflineError("The device is currently offline, no action was sent.")

    def toggle_partymode(self, enabled: bool) -> None:
        """Turn on or off the partymode.

        Args:
            enable (bool): True is enabling partymode, Fasle is disabling partymode.

        Raises:
            NoPartymodeError: Raised if the device does not support partymode.
            OfflineError: Raised if the device is offline.
        """
        if self.online and self.capabilities.check(DeviceCapability.PARTY_MODE):
            if enabled:
                msg = {"sc": {"m": 2, "distm": 0}}
            else:
                msg = {"sc": {"m": 1, "distm": 0}}

            self.mqtt.send(self.name, str(msg))
        elif not self.capabilities.check(DeviceCapability.PARTY_MODE):
            raise NoPartymodeError("This device does not support Partymode")
        elif not self.online:
            raise OfflineError("The device is currently offline, no action was sent.")

    def ots(self, boundary: bool, runtime: str) -> None:
        """Start a One-Time-Schedule task

        Args:
            boundary (bool): If True the device will start the task cutting the edge.
            runtime (str | int): Minutes to run the task before returning to dock.

        Raises:
            NoOneTimeScheduleError: OTS is not supported by the device.
            OfflineError: Raised when the device is offline.
        """
        if self.online and self.capabilities.check(DeviceCapability.ONE_TIME_SCHEDULE):
            if not isinstance(runtime, int):
                runtime = int(runtime)

            msg = {"sc": {"ots": {"bc": int(boundary), "wtm": runtime}}}
            self.mqtt.send(self.name, str(msg))
        elif not self.capabilities.check(DeviceCapability.ONE_TIME_SCHEDULE):
            raise NoOneTimeScheduleError(
                "This device does not support Edgecut-on-demand"
            )
        else:
            raise OfflineError("The device is currently offline, no action was sent.")

    def setzone(self, zone: str) -> None:
        """Set zone to be mowed when next mowing task is started.

        Args:
            zone (str | int): Zone to mow, valid possibilities are a number from 1 to 4.

        Raises:
            OfflineError: Raised if the device is offline.
        """
        if self.online:
            if not isinstance(zone, int):
                zone = int(zone)

            current = self.zone["indicies"]
            new_zones = current
            while not new_zones[self.zone.current] == zone:
                tmp = []
                tmp.append(new_zones[9])
                for i in range(0, 9):
                    tmp.append(new_zones[i])
                new_zones = tmp

            msg = {"mzv": new_zones}
            self.mqtt.send(self.name, str(msg))
        else:
            raise OfflineError("The device is currently offline, no action was sent.")

    def start(self) -> None:
        """Start mowing task

        Raises:
            OfflineError: Raised if the device is offline.
        """
        if self.online:
            logger = _LOGGER.getChild("command")
            logger.debug("Sending START command to %s", self.name)
            self.mqtt.command(self.name, Command.START)
        else:
            raise OfflineError("The device is currently offline, no action was sent.")

    def pause(self) -> None:
        """Pause the mowing task

        Raises:
            OfflineError: Raised if the device is offline.
        """
        if self.online:
            logger = _LOGGER.getChild("command")
            logger.debug("Sending PAUSE command to %s", self.name)
            self.mqtt.command(self.name, Command.PAUSE)
        else:
            raise OfflineError("The device is currently offline, no action was sent.")

    def send(self, data: str) -> None:
        """Send raw JSON data to the device.

        Args:
            data (str): Data to be sent, formatted as a valid JSON object.

        Raises:
            OfflineError: Raised if the device isn't online.
        """
        if self.online:
            logger = _LOGGER.getChild("raw_data_publish")
            logger.debug("Sending %s to %s", data, self.name)
            self.mqtt.send(self.name, data)
        else:
            raise OfflineError("The device is currently offline, no action was sent.")

    def refresh(self) -> None:
        """Try to force a refresh of data from the API."""
        if self.online:
            logger = _LOGGER.getChild("forced_refresh")
            logger.debug("Forcing a data refresh from the API for %s", self.name)
            self.mqtt.send(self.name)
        else:
            raise OfflineError("The device is currently offline, no action was sent.")



"""Handling logger setup."""
from __future__ import annotations

import logging


def get_logger(name: str) -> logging.Logger:
    """Configure the logger component."""

    logger = logging.getLogger(name)

    # configure log formatter
    logFormatter = logging.Formatter(
        "%(asctime)s [%(filename)s] [%(funcName)s] [%(levelname)s] [%(lineno)d] %(message)s"
    )

    # configure stream handler
    consoleHandler = logging.StreamHandler()
    consoleHandler.setFormatter(logFormatter)

    if not len(logger.root.handlers):
        logger.setLevel(logging.DEBUG)
        logger.addHandler(consoleHandler)

    return logger



"""Handler for the physical device information."""
from __future__ import annotations

from enum import IntEnum
from typing import Any

from .landroid_class import LDict


class SetupLocation(LDict):
    """Handling setup location."""

    def __init__(self, latitude: float = 0.0, longitude: float = 0.0) -> None:
        """Initialize setup location object."""
        super().__init__()

        self["latitude"] = latitude
        self["longitude"] = longitude


class WarrantyInfo(LDict):
    """Handling warranty information."""


class InfoType(IntEnum):
    MOWER = 0
    BOARD = 1


class ProductInfo(LDict):
    """Handling mainboard information."""

    def __init__(
        self,
        info_type: InfoType,
        api: Any | None = None,
        product_id: int | None = None,
    ) -> dict:
        """Initialize mower infor object."""
        super().__init__()

        if product_id and api:
            self.get_information_from_id(info_type, api, product_id)

    def get_information_from_id(
        self, info_type: InfoType, api: Any, product_id: int
    ) -> None:
        """Get the device information based on ID."""

        api_prod = None
        if info_type == InfoType.MOWER:
            api_prod = api.get_product_info(product_id)
        elif info_type == InfoType.BOARD:
            api_prod = api.get_board(product_id)

        for attr, val in api_prod.items():
            setattr(self, str(attr), val)

    # def __iter__(self) -> Iterator[dict[str, Any]]:
    #     for key, val in self.items():
    #         if key.startswith("_"):
    #             continue

    #         yield (key, val)

    # def __dict__(self) -> dict[str, Any]:
    #     attrs = {}
    #     for key, val in self.items():
    #         if key.startswith("_"):
    #             continue
    #         attrs.update({key:val})
    #         # yield (key, val)
    #     return attrs



"""Helpers classes."""
from __future__ import annotations

from .logger import get_logger
from .time_format import convert_to_time, string_to_time

__all__ = [convert_to_time, get_logger, string_to_time]



"""Firmware information handler."""
from __future__ import annotations

from typing import Any

from ..const import CONST_UNKNOWN
from .landroid_class import LDict


class Firmware(LDict):
    """Firmware information handler class."""

    def __init__(self, data: Any) -> None:
        super().__init__()

        self["auto_upgrade"] = (
            data["firmware_auto_upgrade"]
            if "firmware_auto_upgrade" in data
            else CONST_UNKNOWN
        )
        self["version"] = (
            data["firmware_version"] if "firmware_version" in data else CONST_UNKNOWN
        )



"""Constants used by Landroid Cloud."""
from __future__ import annotations

API_BASE = "https://{}/api/v2"

UNWANTED_ATTRIBS = [
    "distance_covered",
    "blade_work_time",
    "blade_work_time_reset",
    "blade_work_time_reset_at",
    "battery_charge_cycles",
    "battery_charge_cycles_reset",
    "battery_charge_cycles_reset_at",
    "app_settings",
    "features",
    "iot_registered",
    "pending_radio_link_validation",
    "purchased_at",
    "push_notifications",
    "push_notifications_level",
    "created_at",
    "test",
    "updated_at",
    "warranty_registered",
    "warranty_expires_at",
    "user_id",
    "firmware_auto_upgrade",
    "firmware_version",
    "auto_schedule",
    "auto_schedule_settings",
    "lawn_perimeter",
    "lawn_size",
    "mqtt_topics",
    "mqtt_endpoint",
    "messages_in",
    "messages_out",
    "raw_messages_in",
    "raw_messages_out",
]

CONST_UNKNOWN = "unknown"



"""pyWorxCloud definition."""
from __future__ import annotations

import base64
import contextlib
import json
import sys
import tempfile
from datetime import datetime, timedelta
from typing import Any

import OpenSSL.crypto
import paho.mqtt.client as mqtt
from paho.mqtt.client import connack_string, error_string

from .api import LandroidCloudAPI
from .clouds import CloudType
from .day_map import DAY_MAP
from .events import EventHandler, LandroidEvent
from .exceptions import AuthorizationError, MQTTException
from .helpers import convert_to_time, get_logger
from .utils import (
    MQTT,
    Battery,
    DeviceCapability,
    DeviceHandler,
    Location,
    Orientation,
    ScheduleType,
    Statistic,
    Weekdays,
)
from .utils.schedules import TYPE_TO_STRING

if sys.version_info < (3, 9, 0):
    sys.exit("The pyWorxcloud module requires Python 3.9.0 or later")


class WorxCloud(dict):
    """
    Worx by Landroid Cloud connector.

    Used for handling API connection to Worx, Kress and Landxcape devices which are cloud connected.

    This uses a reverse engineered API protocol, so no guarantee that this will keep working.
    There are no public available API documentation available.
    """

    # __device: str | None = None

    def __init__(
        self,
        username: str,
        password: str,
        cloud: CloudType.WORX
        | CloudType.KRESS
        | CloudType.LANDXCAPE
        | CloudType.FERREX
        | str = CloudType.WORX,
        verify_ssl: bool = True,
    ) -> None:
        """
        Initialize :class:WorxCloud class and set default attribute values.

        1. option for connecting and printing the current states from the API, using :code:`with`

        .. testcode::
        from pyworxcloud import WorxCloud
        from pprint import pprint

        with WorxCloud("your@email","password","worx", 0, False) as cloud:
            pprint(vars(cloud))

        2. option for connecting and printing the current states from the API, using :code:`connect` and :code:`disconnect`

        .. testcode::
        from pyworxcloud import WorxCloud
        from pprint import pprint

        cloud = WorxCloud("your@email", "password", "worx")

        # Initialize connection
        auth = cloud.authenticate()

        if not auth:
            # If invalid credentials are used, or something happend during
            # authorize, then exit
            exit(0)

        # Connect to device with index 0 (devices are enumerated 0, 1, 2 ...)
        # and do not verify SSL (False)
        cloud.connect(0, False)

        # Read latest states received from the device
        cloud.update()

        # Print all vars and attributes of the cloud object
        pprint(vars(cloud))

        # Disconnect from the API
        cloud.disconnect()

        For further information, see the Wiki for documentation: https://github.com/MTrab/pyworxcloud/wiki

        Args:
            username (str): Email used for logging into the app for your device.
            password (str): Password for your account.
            cloud (CloudType.WORX | CloudType.KRESS | CloudType.LANDXCAPE | CloudType.FERREX | str, optional): The CloudType matching your device. Defaults to CloudType.WORX.
            index (int, optional): Device number if more than one is connected to your account (starting from 0 representing the first added device). Defaults to 0.
            verify_ssl (bool, optional): Should this module verify the API endpoint SSL certificate? Defaults to True.

        Raise:
            TypeError: Error raised if invalid CloudType was specified.
        """
        self._worx_mqtt_client_id = None

        if not isinstance(
            cloud,
            (
                type(CloudType.WORX),
                type(CloudType.KRESS),
                type(CloudType.LANDXCAPE),
                type(CloudType.FERREX),
            ),
        ):
            try:
                cloud = getattr(CloudType, cloud.upper())
            except AttributeError:
                raise TypeError(
                    "Wrong type specified, valid types are: worx, landxcape, kress, ferrex"
                )

        self._api = LandroidCloudAPI(username, password, cloud)

        self._username = username
        self._cloud = cloud
        self._auth_result = False
        self._log = get_logger("pyworxcloud")
        self._raw = None

        self._save_zones = None
        self._verify_ssl = verify_ssl
        self._events = EventHandler()

        # Dict of devices, identified by name
        self.devices = {}

        self.mqtt = None

    def __enter__(self) -> Any:
        """Default actions using with statement."""
        self.authenticate()

        self.connect(verify_ssl=self._verify_ssl)
        self.update()

        return self

    def __exit__(self, exc_type, exc_value, traceback) -> Any:
        """Called on end of with statement."""
        self.disconnect()

    def authenticate(self) -> bool:
        """Authenticate against the API."""
        self._log.debug("Authenticating %s", self._username)

        auth = self._authenticate()
        if auth is False:
            self._auth_result = False
            self._log.debug("Authentication for %s failed!", self._username)
            raise AuthorizationError("Unauthorized")

        self._auth_result = True
        self._log.debug("Authentication for %s successful", self._username)

        return True

    def update_attribute(self, device: str, attr: str, key: str, value: Any) -> None:
        """Used as callback to update value."""
        chattr = self.devices[device]
        if not isinstance(attr, type(None)):
            for level in attr.split(";;"):
                if hasattr(chattr, level):
                    chattr = getattr(chattr, level)
                else:
                    chattr = chattr[level]

        if hasattr(chattr, key):
            setattr(chattr, key, value)
        elif isinstance(chattr, dict):
            chattr.update({key: value})

    def set_callback(self, event: LandroidEvent, func: Any) -> None:
        """Set callback which is called when data is received.

        Args:
            event: LandroidEvent for this callback
            func: Function to be called.
        """
        self._events.set_handler(event, func)

    def disconnect(self) -> None:
        """Close API connections."""
        if self.mqtt.connected:
            # topic = self.mqtt.topics["out"]
            # self.mqtt.unsubscribe(topic)
            self.mqtt.disconnect()
            # self.mqtt.loop_stop()

    def connect(
        self,
        verify_ssl: bool = True,
        pahologger: bool = False,
    ) -> bool:
        """Connect to the cloud service endpoint

        Args:
            index (int | None, optional): Device number to connect to. Defaults to None.
            verify_ssl (bool, optional): Should we verify SSL certificate. Defaults to True.

        Returns:
            bool: True if connection was successful, otherwise False.
        """
        self._log.debug("Fetching basic API data")
        self._fetch()
        self._log.debug("Done fetching basic API data")

        self._log.debug("Setting up MQTT handler")
        # setup MQTT handler
        self.mqtt = MQTT(
            self.devices,
            self._worx_mqtt_client_id,
            protocol=mqtt.MQTTv311,
        )
        self._log.debug("Done setting up MQTT handler, setting MQTT config.")

        self.mqtt.endpoint = self._endpoint
        self.mqtt.reconnect_delay_set(60, 300)

        self.mqtt.on_message = self._forward_on_message
        self.mqtt.on_connect = self._on_connect
        self.mqtt.on_disconnect = self._on_disconnect

        if pahologger:
            mqttlog = self._log.getChild("PahoMQTT")
            self.mqtt.on_log = self._on_log
            self.mqtt.enable_logger(mqttlog)
            self.mqtt.logger = True
        self._log.debug("Done setting MQTT config, fetching certificate.")

        with self._get_cert() as cert:
            self.mqtt.tls_set(certfile=cert)
        self._log.debug("Done fetching certificate, setting TLS.")

        if not verify_ssl:
            self.mqtt.tls_insecure_set(True)

        self._log.debug("Done setting TLS, beginning MQTT connect.")

        self.mqtt.connect(self.mqtt.endpoint, port=8883, keepalive=600)
        self._log.debug("MQTT connect done, starting loop")

        self.mqtt.loop_start()
        self._log.debug("MQTT loop started")

        # Convert time strings to objects.

        self._log.debug("Converting date and time string")
        for name, device in self.devices.items():
            convert_to_time(
                name, device, device.time_zone, callback=self.update_attribute
            )
        self._log.debug("Connection tasks all done")

        return True

    @property
    def auth_result(self) -> bool:
        """Return current authentication result."""
        return self._auth_result

    def _authenticate(self):
        """Authenticate the user."""
        auth_data = self._api.auth()

        try:
            self._api.set_token(auth_data["access_token"])
            self._api.set_token_type(auth_data["token_type"])

            self._api.get_profile()
            profile = self._api.data
            if profile is None:
                return False
            self._endpoint = profile["mqtt_endpoint"]
            self._worx_mqtt_client_id = "android-" + self._api.uuid
        except:  # pylint: disable=bare-except
            return False

    @contextlib.contextmanager
    def _get_cert(self):
        """Cet current certificate."""

        certresp = self._api.get_cert()

        with pfx_to_pem(certresp["pkcs12"]) as pem_cert:
            yield pem_cert

    def _forward_on_message(
        self,
        client: mqtt.Client | None,
        userdata: Any | None,
        message: Any | None,
        properties: Any | None = None,  # pylint: disable=unused-argument
    ) -> None:
        """MQTT callback method definition."""
        logger = self._log.getChild("mqtt.message_received")
        topic = message.topic

        for name, topics in self.mqtt.topics.items():
            if topics["out"] == topic:
                break

        device: DeviceHandler = self.devices[name]

        logger.debug(
            "Received MQTT message for %s - processing data %s",
            name,
            message.payload.decode("utf-8"),
        )

        while not device.is_decoded:
            pass  # Await last dataset to be handled before sending a new into the handler

        msg = message.payload.decode("utf-8")
        if device.raw_data == msg:
            self._log.debug("Data was already present and not changed.")
            return  # Data was identical, update was not needed

        device.raw_data = msg
        self._decode_data(device)
        device.capabilities.ready = True

        self._events.call(LandroidEvent.DATA_RECEIVED, name=name, device=device)

    def _decode_data(self, device: DeviceHandler) -> None:
        """Decode incoming JSON data."""
        device.is_decoded = False

        logger = self._log.getChild("decode_data")
        logger.debug("Data decoding for %s started", device.name)

        if device.json_data:
            logger.debug("Found JSON decoded data: %s", device.json_data)
            data = device.json_data
        elif device.raw_data:
            logger.debug("Found raw data: %s", device.raw_data)
            data = json.loads(device.raw_data)
        else:
            device.is_decoded = True
            logger.debug("No valid data was found, skipping update for %s", device.name)
            return

        if "dat" in data:
            device.rssi = data["dat"]["rsi"]
            device.status.update(data["dat"]["ls"])
            device.error.update(data["dat"]["le"])

            device.zone.index = data["dat"]["lz"]

            device.locked = bool(data["dat"]["lk"])

            # Get battery info if available
            if "bt" in data["dat"]:
                if len(device.battery) == 0:
                    device.battery = Battery(data["dat"]["bt"])
                else:
                    device.battery.set_data(data["dat"]["bt"])
            # Get device statistics if available
            if "st" in data["dat"]:
                device.statistics = Statistic(data["dat"]["st"])

            # Get orientation if available.
            if "dmp" in data["dat"]:
                device.orientation = Orientation(data["dat"]["dmp"])

            # Check for extra module availability
            if "modules" in data["dat"]:
                if "4G" in data["dat"]["modules"]:
                    device.gps = Location(
                        data["dat"]["modules"]["4G"]["gps"]["coo"][0],
                        data["dat"]["modules"]["4G"]["gps"]["coo"][1],
                    )

            # Get remaining rain delay if available
            if "rain" in data["dat"]:
                device.rainsensor.triggered = bool(str(data["dat"]["rain"]["s"]) == "1")
                device.rainsensor.remaining = int(data["dat"]["rain"]["cnt"])

        if "cfg" in data:
            device.updated = data["cfg"]["dt"] + " " + data["cfg"]["tm"]
            device.rainsensor.delay = int(data["cfg"]["rd"])

            # Fetch wheel torque
            if "tq" in data["cfg"]:
                device.capabilities.add(DeviceCapability.TORQUE)
                device.torque = data["cfg"]["tq"]

            # Fetch zone information
            if "mz" in data["cfg"]:
                device.zone.starting_point = data["cfg"]["mz"]
                device.zone.indicies = data["cfg"]["mzv"]

                # Map current zone to zone index
                device.zone.current = device.zone.indicies[device.zone.index]

            # Fetch main schedule
            if "sc" in data["cfg"]:
                if "ots" in data["cfg"]["sc"]:
                    device.capabilities.add(DeviceCapability.ONE_TIME_SCHEDULE)
                    device.capabilities.add(DeviceCapability.EDGE_CUT)
                if "distm" in data["cfg"]["sc"]:
                    device.capabilities.add(DeviceCapability.PARTY_MODE)

                device.partymode_enabled = bool(str(data["cfg"]["sc"]["m"]) == "2")

                device.schedules["active"] = bool(str(data["cfg"]["sc"]["m"]) == "1")
                device.schedules["time_extension"] = data["cfg"]["sc"]["p"]

                sch_type = ScheduleType.PRIMARY
                device.schedules.update({TYPE_TO_STRING[sch_type]: Weekdays()})

                for day in range(0, len(data["cfg"]["sc"]["d"])):
                    device.schedules[TYPE_TO_STRING[sch_type]][DAY_MAP[day]][
                        "start"
                    ] = data["cfg"]["sc"]["d"][day][0]
                    device.schedules[TYPE_TO_STRING[sch_type]][DAY_MAP[day]][
                        "duration"
                    ] = data["cfg"]["sc"]["d"][day][1]
                    device.schedules[TYPE_TO_STRING[sch_type]][DAY_MAP[day]][
                        "boundary"
                    ] = bool(data["cfg"]["sc"]["d"][day][2])

                    time_start = datetime.strptime(
                        device.schedules[TYPE_TO_STRING[sch_type]][DAY_MAP[day]][
                            "start"
                        ],
                        "%H:%M",
                    )

                    if isinstance(
                        device.schedules[TYPE_TO_STRING[sch_type]][DAY_MAP[day]][
                            "duration"
                        ],
                        type(None),
                    ):
                        device.schedules[TYPE_TO_STRING[sch_type]][DAY_MAP[day]][
                            "duration"
                        ] = "0"

                    duration = int(
                        device.schedules[TYPE_TO_STRING[sch_type]][DAY_MAP[day]][
                            "duration"
                        ]
                    )

                    duration = duration * (
                        1 + (int(device.schedules["time_extension"]) / 100)
                    )
                    end_time = time_start + timedelta(minutes=duration)

                    device.schedules[TYPE_TO_STRING[sch_type]][DAY_MAP[day]][
                        "end"
                    ] = end_time.time().strftime("%H:%M")

            # Fetch secondary schedule
            if "dd" in data["cfg"]["sc"]:
                sch_type = ScheduleType.SECONDARY
                device.schedules.update({TYPE_TO_STRING[sch_type]: Weekdays()})

                for day in range(0, len(data["cfg"]["sc"]["d"])):
                    device.schedules[TYPE_TO_STRING[sch_type]][DAY_MAP[day]][
                        "start"
                    ] = data["cfg"]["sc"]["dd"][day][0]
                    device.schedules[TYPE_TO_STRING[sch_type]][DAY_MAP[day]][
                        "duration"
                    ] = data["cfg"]["sc"]["dd"][day][1]
                    device.schedules[TYPE_TO_STRING[sch_type]][DAY_MAP[day]][
                        "boundary"
                    ] = bool(data["cfg"]["sc"]["dd"][day][2])

                    time_start = datetime.strptime(
                        data["cfg"]["sc"]["dd"][day][0],
                        "%H:%M",
                    )

                    if isinstance(
                        device.schedules[TYPE_TO_STRING[sch_type]][DAY_MAP[day]][
                            "duration"
                        ],
                        type(None),
                    ):
                        device.schedules[TYPE_TO_STRING[sch_type]][DAY_MAP[day]][
                            "duration"
                        ] = "0"

                    duration = int(
                        device.schedules[TYPE_TO_STRING[sch_type]][DAY_MAP[day]][
                            "duration"
                        ]
                    )

                    duration = duration * (
                        1 + (int(device.schedules["time_extension"]) / 100)
                    )
                    end_time = time_start + timedelta(minutes=duration)

                    device.schedules[TYPE_TO_STRING[sch_type]][DAY_MAP[day]][
                        "end"
                    ] = end_time.time().strftime("%H:%M")

        convert_to_time(
            device.name, device, device.time_zone, callback=self.update_attribute
        )

        device.is_decoded = True
        logger.debug("Data for %s was decoded", device.name)

    def _on_log(
        self,
        client: mqtt.Client | None,
        userdata: Any | None,
        level: Any | None,
        buf: Any | None,
    ) -> None:
        """Capture MQTT log messages."""
        logger = self._log.getChild("mqtt.log")
        logger.debug("MQTT log message received: %s", buf)

    def _on_connect(
        self,
        client: mqtt.Client | None,
        userdata: Any | None,
        flags: Any | None,
        rc: int | None,
        properties: Any | None = None,  # pylint: disable=unused-argument,invalid-name
    ) -> None:
        """MQTT callback method."""
        logger = self._log.getChild("mqtt.connected")
        logger.debug(connack_string(rc))
        if rc == 0:
            for name, topics in self.mqtt.topics.items():
                topic = topics["out"]
                logger.debug(
                    "MQTT for %s connected, subscribing to topic '%s'", name, topic
                )
                client.subscribe(topic)

            for name, device in self.devices.items():
                device.mqtt = self.mqtt
                if isinstance(device.raw_data, type(None)):
                    logger.debug(
                        "MQTT chached data not found for %s - requesting now", name
                    )

                    mqp = device.mqtt.send(name, force=True)
                    if isinstance(mqp, type(None)):
                        raise MQTTException("Couldn't send request to MQTT server.")
                    elif not isinstance(mqp, str):
                        while not mqp.is_published:
                            pass

            logger.debug("Setting MQTT connected flag TRUE")
            self.mqtt.connected = True
            self._events.call(LandroidEvent.MQTT_CONNECTION, state=self.mqtt.connected)

        else:
            logger.debug("Setting MQTT connected flag FALSE")
            self.mqtt.connected = False
            self._events.call(LandroidEvent.MQTT_CONNECTION, state=self.mqtt.connected)

            raise MQTTException(connack_string(rc))

    def _on_disconnect(
        self,
        client: mqtt.Client | None,
        userdata: Any | None,
        rc: int | None,
        properties: Any | None = None,  # pylint: disable=unused-argument,invalid-name
    ) -> None:
        """MQTT callback method."""
        logger = self._log.getChild("mqtt.disconnected")
        if rc > 0:
            if rc == 7:
                if not self.mqtt.connected:
                    raise MQTTException(
                        "Unexpected MQTT disconnect - were you perhaps banned?"
                    )

            if self.mqtt.connected:
                logger.debug("MQTT connection was lost! (%s)", error_string(rc))

            logger.debug("Setting MQTT connected flag FALSE")
            self.mqtt.connected = False
            self._events.call(LandroidEvent.MQTT_CONNECTION, state=self.mqtt.connected)
            for name, topics in self.mqtt.topics.items():
                topic = topics["out"]
                logger.debug(
                    "MQTT for %s disconnected, unsubscribing from topic '%s'",
                    name,
                    topic,
                )
                client.unsubscribe(topic)

    def _fetch(self) -> None:
        """Fetch base API information."""
        products = self._api.get_products()

        for product in products:
            device = DeviceHandler(self._api, product)
            self.devices.update({product["name"]: device})

    def update(self) -> None:
        """Retrive current device status."""
        for name, device in self.devices.items():
            status = self._api.get_status(device.serial_number)
            status = str(status).replace("'", '"')

            while not device.is_decoded:
                pass  # Await previous dataset to be handled before sending a new into the handler.

            device.raw_data = status
            self._decode_data(device)


@contextlib.contextmanager
def pfx_to_pem(pfx_data):
    """Decrypts the .pfx file to be used with requests."""
    with tempfile.NamedTemporaryFile(suffix=".pem", delete=False) as t_pem:
        f_pem = open(t_pem.name, "wb")
        p12 = OpenSSL.crypto.load_pkcs12(base64.b64decode(pfx_data), "")
        f_pem.write(
            OpenSSL.crypto.dump_privatekey(
                OpenSSL.crypto.FILETYPE_PEM, p12.get_privatekey()
            )
        )
        f_pem.write(
            OpenSSL.crypto.dump_certificate(
                OpenSSL.crypto.FILETYPE_PEM, p12.get_certificate()
            )
        )
        certauth = p12.get_ca_certificates()
        if certauth is not None:
            for cert in certauth:
                f_pem.write(
                    OpenSSL.crypto.dump_certificate(OpenSSL.crypto.FILETYPE_PEM, cert)
                )
        f_pem.close()
        yield t_pem.name



"""Map days to integer."""
from __future__ import annotations

DAY_MAP = {
    0: "sunday",
    1: "monday",
    2: "tuesday",
    3: "wednesday",
    4: "thursday",
    5: "friday",
    6: "saturday",
}



"""MQTT information class."""
from __future__ import annotations

import asyncio
import math
import re
from datetime import datetime, timedelta
from typing import Any

import paho.mqtt.client as mqtt
from paho.mqtt.client import MQTTMessageInfo
from ratelimit import RateLimitException, limits

from ..events import EventHandler, LandroidEvent
from ..exceptions import MQTTException
from ..helpers import get_logger
from .landroid_class import LDict

_LOGGER = get_logger("mqtt")

MQTT_IN = "{}/{}/commandIn"
MQTT_OUT = "{}/{}/commandOut"

PUBLISH_LIMIT_PERIOD = 60  # s
PUBLISH_CALLS_LIMIT = 5  # polls per timeframe


class MQTTMsgType(LDict):
    """Define specific message type data."""

    def __init__(self) -> dict:
        super().__init__()

        self["in"] = 0
        self["out"] = 0


class MQTTMessageItem(LDict):
    """Defines a MQTT message for Landroid Cloud."""

    def __init__(
        self, device: str, data: str = "{}", qos: int = 0, retain: bool = False
    ) -> dict:
        super().__init__()

        self["device"] = device
        self["data"] = data
        self["qos"] = qos
        self["retain"] = retain


class MQTTQueue:
    """Class for handling queue."""

    def __init__(self):
        """Initialize queue object."""
        self.callback: Any | None = None
        self.retry_at = datetime.now()
        self.items = list[MQTTMessageItem]()


class MQTTMessages(LDict):
    """Messages class."""

    def __init__(self) -> dict:
        super().__init__()

        self["raw"] = MQTTMsgType()
        self["filtered"] = MQTTMsgType()


class MQTTTopics(LDict):
    """Topics class."""

    def __init__(
        self, topic_in: str | None = None, topic_out: str | None = None
    ) -> dict:
        super().__init__()

        self["in"] = topic_in
        self["out"] = topic_out


class Command:
    """Landroid Cloud commands."""

    START = 1
    PAUSE = 2
    HOME = 3
    ZONETRAINING = 4
    LOCK = 5
    UNLOCK = 6
    RESTART = 7
    PAUSE_OVER_WIRE = 8
    SAFEHOME = 9


class MQTT(mqtt.Client, LDict):
    """Full MQTT handler class."""

    __loop: asyncio.AbstractEventLoop | None = None
    __loop_allow: bool = True

    def __init__(
        self,
        devices: Any | None = None,
        client_id: str | None = None,
        clean_session: Any | None = None,
        userdata: Any | None = None,
        protocol: Any = mqtt.MQTTv311,
        transport: str = "tcp",
        reconnect_on_failure: bool = True,
    ) -> dict:
        if isinstance(devices, type(None)):
            return

        super().__init__(
            client_id,
            clean_session,
            userdata,
            protocol,
            transport,
            reconnect_on_failure,
        )

        self._events = EventHandler()

        self.devices = devices

        self.endpoint = None
        self.connected = False

        self.queue = MQTTQueue()
        self.topics = {}
        for name, device in devices.items():
            topic_in = MQTT_IN.format(device.mainboard.code, device.mac_address)
            topic_out = MQTT_OUT.format(device.mainboard.code, device.mac_address)
            self.topics.update({name: MQTTTopics(topic_in, topic_out)})

        # try:
        #     self.__loop = asyncio.get_event_loop()
        #     self.__loop.run_in_executor(None, self.__handle_queue)
        # except RuntimeError:
        #     return

    def set_eventloop(self, eventloop: Any) -> None:
        """Set eventloop to be used ny queue handler."""
        self.__loop = eventloop
        self.__loop.run_in_executor(None, self.__handle_queue)

    def stop_queuehandler(self) -> None:
        """Stops the eventloop for the queue handler."""
        self.queue.callback = None
        self.__loop_allow = False

    def disconnect(self, reasoncode=None, properties=None):
        self.__loop_allow = False
        self.loop_stop()
        return super().disconnect(reasoncode, properties)

    def __handle_queue(self):
        """Handle the MQTT queue."""
        while self.__loop_allow:
            if isinstance(self.queue.retry_at, int):
                # await asyncio.sleep(0.01)
                continue

            if self.queue.retry_at < datetime.now() and len(self.queue.items) > 0:
                queue_list = list(reversed(self.queue.items))
                self.queue.items.clear()
                while queue_list:
                    message = queue_list.pop()
                    log_msg = f'Trying message "{message}" from the message queue'
                    if not self._events.call(
                        LandroidEvent.MQTT_RATELIMIT, message=log_msg
                    ):
                        _LOGGER.debug(log_msg)
                    self.send(
                        device=message["device"],
                        data=message["data"],
                        qos=message["qos"],
                        retain=message["retain"],
                    )

    @limits(calls=PUBLISH_CALLS_LIMIT, period=PUBLISH_LIMIT_PERIOD)
    def __send(
        self,
        topic: str,
        data: str = "{}",
        qos: int = 0,
        retain: bool = False,
    ) -> MQTTMessageInfo:
        """Do the actual publish."""
        return self.publish(topic, data, qos, retain)

    def send(
        self,
        device: str,
        data: str = "{}",
        qos: int = 0,
        retain: bool = False,
        force: bool = False,
    ) -> MQTTMessageInfo | str:
        """Send Landroid cloud message to API endpoint."""
        from .devices import DeviceHandler

        recipient: DeviceHandler = self.devices[device]
        topic = self.topics[device]["in"]

        data = data.replace("'", '"')
        if not re.match('^\{[A-ZÆØÅa-zæøå0-9:,\[\]"{} \n]*\}$', data):
            data = "{" + data + "}"

        log_msg = f'Sending "{data}" to "{recipient.name}" on "{topic}"'
        if not self._events.call(LandroidEvent.LOG, message=log_msg, level="debug"):
            _LOGGER.debug(log_msg)

        if not self.connected and not force:
            log_msg = (
                f"MQTT server was not connected, can"
                't send message to "{recipient.name}"'
            )
            if not self._events.call(LandroidEvent.LOG, message=log_msg, level="error"):
                _LOGGER.error(log_msg)

            raise MQTTException("MQTT not connected")

        message = MQTTMessageItem(device, data, qos, retain)

        self._events.call(
            LandroidEvent.MQTT_PUBLISH,
            message=data,
            device=device,
            topic=topic,
            qos=qos,
            retain=retain,
        )

        try:
            status = self.__send(topic, data, qos, retain)
            log_msg = (
                f'Awaiting message to be published to "{recipient.name}" on "{topic}"'
            )
            if not self._events.call(LandroidEvent.LOG, message=log_msg, level="debug"):
                _LOGGER.debug(log_msg)

            while not status.is_published:
                pass  # Await status to change to is_published

            log_msg = f'MQTT message was published to "{recipient.name}" on "{topic}"'
            if not self._events.call(LandroidEvent.LOG, message=log_msg, level="debug"):
                _LOGGER.debug(log_msg)

            return status
        except ValueError as exc:
            log_msg = f'MQTT queue for "{recipient.name}" was full, message "{data}" was not sent!'
            if not self._events.call(LandroidEvent.LOG, message=log_msg, level="error"):
                _LOGGER.error(log_msg)
        except RuntimeError as exc:
            log_msg = f'MQTT error while sending message "{data}" to "{recipient.name}"\n{exc}'
            if not self._events.call(LandroidEvent.LOG, message=log_msg, level="error"):
                _LOGGER.error(log_msg)
        except RateLimitException as exc:
            _LOGGER.debug("Adding '%s' to message queue.", message)
            self.queue.retry_at = datetime.now() + timedelta(
                seconds=math.ceil(exc.period_remaining)
            )
            self.queue.items.append(message)
            self._events.call(
                LandroidEvent.MQTT_RATELIMIT,
                message=f"Ratelimit of {PUBLISH_CALLS_LIMIT} messages in {PUBLISH_LIMIT_PERIOD} seconds exceeded. Message '{message['data']}' to '{message['device']}' added to message queue.",
            )
            return f"Ratelimit of {PUBLISH_CALLS_LIMIT} messages in {PUBLISH_LIMIT_PERIOD} seconds exceeded. Wait {math.ceil(exc.period_remaining)} seconds before trying again"
        except Exception as exc:
            log_msg = f'MQTT error sending "{data}" to "{recipient.name}"'
            if not self._events.call(LandroidEvent.LOG, message=log_msg, level="error"):
                _LOGGER.error(log_msg)

    def command(self, device: str, action: Command) -> MQTTMessageInfo:
        """Send command to device."""
        cmd = '"cmd":{}'.format(action)
        cmd = "{" + cmd + "}"
        return self.send(device, cmd)



"""Class for handling device info and states."""
from __future__ import annotations

import json
from typing import Any

from ..const import UNWANTED_ATTRIBS
from ..exceptions import APIException
from .actions import Actions
from .battery import Battery
from .blades import Blades
from .capability import Capability
from .firmware import Firmware
from .landroid_class import LDict
from .lawn import Lawn
from .orientation import Orientation
from .product import InfoType, ProductInfo
from .rainsensor import Rainsensor
from .schedules import Schedule
from .state import States, StateType
from .warranty import Warranty
from .zone import Zone


class DeviceHandler(LDict, Actions):
    """DeviceHandler for Landroid Cloud devices."""

    __is_decoded: bool = True
    __raw_data: str = None
    __json_data: str = None

    def __init__(self, api: Any = None, product: Any = None) -> dict:
        """Initialize the object."""
        super().__init__()

        if not isinstance(product, type(None)) and not isinstance(api, type(None)):
            self.__mapinfo(api, product)

    @property
    def raw_data(self) -> str:
        """Returns current raw dataset."""
        return self.__raw_data

    @property
    def json_data(self) -> str:
        """Returns current dataset as JSON."""
        return self.__json_data

    @raw_data.setter
    def raw_data(self, value: str) -> None:
        """Set new MQTT data."""
        self.__is_decoded = False
        self.__raw_data = value
        try:
            self.__json_data = json.loads(value)
        except:
            pass  # Just continue if we couldn't decode the data

    @property
    def is_decoded(self) -> bool:
        """Returns true if latest dataset was decoded and handled."""
        return self.__is_decoded

    @is_decoded.setter
    def is_decoded(self, value: bool) -> None:
        """Set decoded flag when dataset was decoded and handled."""
        self.__is_decoded = value

    def __mapinfo(self, api: Any, data: Any) -> None:
        """Map information from API."""

        if isinstance(data, type(None)) or isinstance(api, type(None)):
            raise APIException(
                "Either 'data' or 'api' object was missing, no data was mapped!"
            )

        for attr, val in data.items():
            setattr(self, str(attr), val)

        self.battery = Battery(data)
        self.blades = Blades(data)
        self.chassis = ProductInfo(InfoType.MOWER, api, data["product_id"])
        self.mainboard = ProductInfo(InfoType.BOARD, api, self.chassis.board_id)
        self.error = States(StateType.ERROR)
        self.orientation = Orientation([0, 0, 0])
        self.capabilities = Capability()
        self.rainsensor = Rainsensor()
        self.status = States()
        self.zone = Zone()
        self.warranty = Warranty(data)
        self.firmware = Firmware(data)
        self.schedules = Schedule()

        if data in ["lawn_perimeter", "lawn_size"]:
            self.lawn = Lawn(data["lawn_perimeter"], data["lawn_size"])

        if data in ["auto_schedule_settings", "auto_schedule"]:
            self.schedules["auto_schedule"]["settings"] = data["auto_schedule_settings"]
            self.schedules["auto_schedule"]["enabled"] = data["auto_schedule"]

        self.name = data["name"]
        self.model = f"{self.chassis.default_name}{self.chassis.meters}"

        for attr in UNWANTED_ATTRIBS:
            if hasattr(self, attr):
                delattr(self, attr)

        self.is_decoded = True



"""A class set that helps representing data as wanted."""
from __future__ import annotations

from typing import Any


class LDict(dict):
    """A Landroid custom dict."""

    def __init__(self, default: Any | None = None):
        """Init dict."""
        super().__init__(self)



"""Rain sensor information."""
from __future__ import annotations

from .landroid_class import LDict


class Rainsensor(LDict):
    """Rain sensor definition."""

    def __init__(self) -> dict:
        """Initialize rain sensor object."""
        super().__init__()

        self["delay"] = 0

    @property
    def delay(self) -> int:
        """Return rain delay."""
        return self["delay"]

    @delay.setter
    def delay(self, raindelay: int) -> None:
        """Set rain delay information."""
        self["delay"] = raindelay

    @property
    def triggered(self) -> bool:
        """Return rain sensor trigger state."""
        return self["triggered"]

    @triggered.setter
    def triggered(self, triggered: bool) -> None:
        """Set rain sensor trigger state."""
        self["triggered"] = triggered

    @property
    def remaining(self) -> int:
        """Return remaining rain delay."""
        return self["remaining"]

    @remaining.setter
    def remaining(self, remaining: int) -> None:
        """Set remaining rain delay information."""
        self["remaining"] = remaining



"""Utils."""
from __future__ import annotations

from .battery import Battery
from .blades import Blades
from .capability import Capability, DeviceCapability
from .devices import DeviceHandler
from .location import Location
from .mqtt import MQTT, Command
from .orientation import Orientation
from .product import ProductInfo
from .rainsensor import Rainsensor
from .schedules import Schedule, ScheduleType, Weekdays
from .state import States, StateType
from .statistics import Statistic
from .warranty import Warranty
from .zone import Zone

__all__ = [
    Battery,
    Blades,
    Capability,
    Command,
    DeviceHandler,
    DeviceCapability,
    Location,
    MQTT,
    Orientation,
    ProductInfo,
    Rainsensor,
    Schedule,
    ScheduleType,
    States,
    StateType,
    Statistic,
    Warranty,
    Weekdays,
    Zone,
]



"""Landroid Cloud statistics class."""
from __future__ import annotations

from .landroid_class import LDict


class Statistic(LDict):
    """Statistics."""

    def __init__(self, data: list | None = None) -> dict:
        """Initialize a statistics class."""
        super().__init__()

        if not data:
            return

        # Total runtime with blades on in minutes
        self["worktime_blades_on"] = data["b"] if "b" in data else None

        # Total distance in meters
        self["distance"] = data["d"] if "d" in data else None

        # Total worktime in minutes
        self["worktime_total"] = data["wt"] if "wt" in data else None



"""Defines schedule classes."""
from __future__ import annotations

import calendar
from enum import IntEnum

from .landroid_class import LDict


class ScheduleType(IntEnum):
    """Schedule types."""

    PRIMARY = 0
    SECONDARY = 1


TYPE_TO_STRING = {ScheduleType.PRIMARY: "primary", ScheduleType.SECONDARY: "secondary"}


class WeekdaySettings(LDict):
    """Class representing a weekday setting."""

    def __init__(
        self,
        start: str = "00:00",
        end: str = "00:00",
        duration: int = 0,
        boundary: bool = False,
    ) -> None:
        """Initialize the settings."""
        super().__init__()
        self["start"] = start
        self["end"] = end
        self["duration"] = duration
        self["boundary"] = boundary


class Weekdays(LDict):
    """Represents all weekdays."""

    def __init__(self) -> dict:
        super().__init__()

        for day in list(calendar.day_name):
            self.update({day.lower(): WeekdaySettings()})


class Schedule(LDict):
    """Represents a schedule."""

    def __init__(
        self,
        variation: int = 0,
        active: bool = True,
        auto_schedule_settings: dict = {},
        auto_schedule_enabled: bool | None = None,
    ) -> dict:
        """Initialize an empty primary or secondary schedule.

        Args:
            schedule_type (ScheduleType): Which ScheduleType to initialize.
        """
        super().__init__()

        self["time_extension"] = variation
        self["active"] = active
        self["auto_schedule"] = {
            "settings": auto_schedule_settings,
            "enabled": auto_schedule_enabled,
        }



"""Landroid Cloud API implementation"""
# pylint: disable=unnecessary-lambda
from __future__ import annotations

import functools
import json
import operator
import re
import time
import uuid
from typing import Any

import requests

from .clouds import CloudType
from .const import API_BASE
from .exceptions import (
    APIException,
    AuthorizationError,
    ForbiddenError,
    InternalServerError,
    NotFoundError,
    RequestError,
    RequestException,
    ServiceUnavailableError,
    TimeoutException,
    TokenError,
    TooManyRequestsError,
)


class LandroidCloudAPI:
    """Landroid Cloud API definition."""

    def __init__(
        self,
        username: str,
        password: str,
        cloud: CloudType.WORX
        | CloudType.KRESS
        | CloudType.LANDXCAPE
        | CloudType.FERREX = CloudType.WORX,
    ) -> None:
        """Initialize a new instance of the API broker.

        Args:
            username (str): Email for the user account.
            password (str): Password for the user account.
            cloud (CloudType.WORX | CloudType.KRESS | CloudType.LANDXCAPE | CloudType.FERREX, optional): CloudType representing the device. Defaults to CloudType.WORX.
        """
        self.cloud: CloudType = cloud
        self._token_type = "app"
        self._token = None
        self._tokenrefresh = 0
        self.uuid = None
        self._api_host = None
        self._data = None

        self.api_url = cloud.URL
        self.api_key = cloud.KEY
        self.username = username
        self.password = password

    def set_token(self, token: str, now: int = -1) -> None:
        """Set the API token to be used

        Args:
            token (str): Token from authentication.
            now (int, optional): When was the token last refreshed. Defaults to -1 which will force a token refresh.
        """
        if now == -1:
            now = int(time.time())

        self._tokenrefresh = now
        self._token = token

    def set_token_type(self, token_type: str) -> None:
        """Set token type.

        Args:
            token_type (str): Token type
        """
        self._token_type = token_type

    def _generate_identify_token(self, seed_token: str) -> str:
        """Generate identify token."""
        text_to_char = [ord(c) for c in self.api_url]

        step_one = re.findall(r".{1,2}", seed_token)
        step_two = list(map((lambda hex: int(hex, 16)), step_one))

        step_three = list(
            map(
                (
                    lambda foo: functools.reduce(
                        (lambda x, y: operator.xor(x, y)),
                        text_to_char,
                        foo,
                    )
                ),
                step_two,
            )
        )
        step_four = list(map(chr, step_three))

        final = "".join(step_four)
        return final

    def _get_headers(self) -> dict:
        """Create header object for communication packets."""
        header_data = {}
        header_data["Content-Type"] = "application/json"
        header_data["Authorization"] = self._token_type + " " + self._token

        return header_data

    def auth(self) -> str:
        """Authenticate against API endpoint

        Returns:
            str: JSON object containing authentication token.
        """
        self.uuid = str(uuid.uuid1())
        self._api_host = (API_BASE).format(self.api_url)

        self._token = self._generate_identify_token(self.api_key)

        payload_data = {}
        payload_data["username"] = self.username
        payload_data["password"] = self.password
        payload_data["grant_type"] = "password"
        payload_data["client_id"] = 1
        payload_data["type"] = "app"
        payload_data["client_secret"] = self._token
        payload_data["scope"] = "*"

        payload = json.dumps(payload_data)

        calldata = self._call("/oauth/token", payload, checktoken=False)

        return calldata

    def get_profile(self) -> str:
        """Get user profile.

        Returns:
            str: JSON object with user data.
        """
        calldata = self._call("/users/me")
        self._data = calldata
        return calldata

    def get_cert(self) -> str:
        """Get the user certificate.

        Returns:
            str: JSON object with certificate info.
        """
        calldata = self._call("/users/certificate")
        self._data = calldata
        return calldata

    def get_products(self) -> str:
        """Get products associated with the account.

        Returns:
            str: JSON object containing available devices associated with the account.
        """
        calldata = self._call("/product-items")
        self._data = calldata
        return calldata

    def get_product_info(self, product_id: int) -> dict[str, Any]:
        """Get product info and features for a given device model.

        Args:
            product_id (int): Product_id to get information for.

        Returns:
            list | None: A list of attributes with the product information.
        """
        products = self._call("/products")
        try:
            return [x for x in products if x["id"] == product_id][0]
        except:
            return None

    def get_board(self, board_id: int) -> list:
        """Get board info and features for a given baseboard.

        Args:
            board_code (str): Board_id to get information for.

        Returns:
            list | None: A list of attributes with board information.
        """
        boards = self._call("/boards")
        try:
            return [x for x in boards if x["id"] == board_id][0]
        except:
            return None

    def get_status(self, serial: str) -> str:
        """Get device status

        Args:
            serial (str): Serialnumber for the device to get status for.

        Returns:
            str: JSON object containing the device status.
        """
        callstr = f"/product-items/{serial}/status"
        calldata = self._call(callstr)
        return calldata

    def _call(
        self, path: str, payload: str | None = None, checktoken: bool = True
    ) -> str:
        """Do the actual call to the device."""
        # Check if token needs refreshing
        now = int(time.time())  # Current time in unix timestamp format
        if checktoken and ((self._tokenrefresh + 3000) < now):
            try:
                auth_data = self.auth()

                if "return_code" in auth_data:
                    return

                self.set_token(auth_data["access_token"], now)
                self.set_token_type(auth_data["token_type"])
            except Exception as ex:  # pylint: disable=bare-except
                raise TokenError("Error refreshing authentication token") from ex

        try:
            if payload:
                req = requests.post(
                    self._api_host + path,
                    data=payload,
                    headers=self._get_headers(),
                    timeout=30,
                )
            else:
                req = requests.get(
                    self._api_host + path, headers=self._get_headers(), timeout=30
                )

            req.raise_for_status()
        except requests.exceptions.HTTPError as err:
            code = err.response.status_code
            if code == 400:
                raise RequestError()
            elif code == 401:
                raise AuthorizationError()
            elif code == 403:
                raise ForbiddenError()
            elif code == 404:
                raise NotFoundError()
            elif code == 429:
                raise TooManyRequestsError()
            elif code == 500:
                raise InternalServerError()
            elif code == 503:
                raise ServiceUnavailableError()
            else:
                raise APIException(err)
        except requests.exceptions.Timeout as err:
            raise TimeoutException(err)
        except requests.exceptions.RequestException as err:
            raise RequestException(err)

        return req.json()

    @property
    def data(self) -> str:
        """Return the latest dataset of information and states from the API."""
        return self._data



"""Battery information."""
from __future__ import annotations

from enum import IntEnum
from typing import Any

from .landroid_class import LDict


class BatteryState(IntEnum):
    """Battery states."""

    UNKNOWN = -1
    NOT_CHARGING = 0
    CHARGING = 1
    ERROR_CHARGING = 2


CHARGE_MAP = {
    BatteryState.UNKNOWN: "unknown",
    BatteryState.NOT_CHARGING: False,
    BatteryState.CHARGING: True,
    BatteryState.ERROR_CHARGING: "error",
}


class Battery(LDict):
    """Battery information."""

    def __init__(
        self, indata: list | None = None, cycle_info: Any | None = None
    ) -> None:
        """Initialize a battery object."""
        super().__init__()

        if not indata and not cycle_info:
            return

        if not "cycles" in self:
            self["cycles"] = {
                "total": 0,
                "current": 0,
                "reset_at": None,
                "reset_time": None,
            }

        if indata:
            self.set_data(indata)
            self._update_cycles()

        if cycle_info:
            self._set_cycles(cycle_info)

    def set_data(self, indata: list):
        """Update data on existing dataset."""
        if "t" in indata:
            self["temperature"] = indata["t"]
        if "v" in indata:
            self["voltage"] = indata["v"]
        if "p" in indata:
            self["percent"] = indata["p"]
        if "c" in indata:
            self["charging"] = CHARGE_MAP[indata["c"]]
        if "nr" in indata:
            self["cycles"].update({"total": indata["nr"]})

    def _update_cycles(self) -> None:
        """Update cycles info."""
        if self["cycles"]["total"] == 0:
            return
        elif (
            isinstance(self["cycles"]["reset_at"], type(None))
            and self["cycles"]["total"] > 0
        ):
            self["cycles"].update({"current": self["cycles"]["total"]})
        else:
            self["cycles"].update(
                {"current": int(self["cycles"]["total"] - self["cycles"]["reset_at"])}
            )

    def _set_cycles(self, indata) -> None:
        """Set battery cycles information."""
        from ..helpers import string_to_time

        if self["cycles"]["total"] == 0:
            self["cycles"].update({"total": indata.battery_charge_cycles})

        if indata.battery_charge_cycles_reset is not None:
            if self["cycles"]["total"] == 0:
                self["cycles"].update(
                    {
                        "current": int(
                            self["cycles"]["total"] - indata.battery_charge_cycles_reset
                        )
                    }
                )
                if self["cycles"]["current"] < 0:
                    self["cycles"].update({"current": 0})
            self["cycles"].update(
                {
                    "reset_at": int(indata.battery_charge_cycles_reset),
                    "reset_time": string_to_time(
                        indata.battery_charge_cycles_reset_at, indata.time_zone
                    )
                    if not isinstance(indata.battery_charge_cycles_reset_at, type(None))
                    else None,
                }
            )
        else:
            if self["cycles"]["total"] > 0:
                self["cycles"].update({"current": self["cycles"]["total"]})



"""Zone representation."""
from __future__ import annotations

from .landroid_class import LDict


class Zone(LDict):
    """Class for handling zone data."""

    def __init__(self) -> dict:
        """Initialize zone object."""
        super().__init__()

        self["current"] = None
        self["index"] = None
        self["indicies"] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        self["starting_point"] = [0, 0, 0, 0]

    @property
    def current(self) -> int:
        """Get current zone."""
        return self["current"]

    @current.setter
    def current(self, value: int) -> None:
        """Set current zone property."""
        self["current"] = value

    @property
    def index(self) -> int:
        """Get current index."""
        return self["index"]

    @index.setter
    def index(self, value: int) -> None:
        """Set current index property."""
        self["index"] = value

    @property
    def indicies(self) -> int:
        """Get indicies."""
        return self["indicies"]

    @indicies.setter
    def indicies(self, value: list) -> None:
        """Set indicies property."""
        self["indicies"] = value

    @property
    def starting_point(self) -> int:
        """Get starting points."""
        return self["starting_point"]

    @starting_point.setter
    def starting_point(self, value: int) -> None:
        """Set starting points."""
        self["starting_point"] = value



"""Time formatting helpers."""
from __future__ import annotations

import re
from datetime import datetime
from typing import Any

import pytz

try:
    from ..utils import __all__ as all_utils
except:
    pass

DATE_FORMATS = [
    "%Y-%m-%d %H:%M:%S",
    "%d-%m-%Y %H:%M:%S",
    "%Y/%m/%d %H:%M:%S",
    "%d/%m/%Y %H:%M:%S",
]


def string_to_time(dt_string: str, tz: str = "UTC") -> datetime | str:
    """Convert string to datetime object.
    Trying all known date/time formats as defined in DATE_FORMATS constant.

    Args:
        dt_string (str): String containing the date/time
        tz (str): Timezone for the string. default = "UTC"

    Returns:
        datetime: datatime object
    """
    timezone = (
        pytz.timezone(tz) if not isinstance(tz, type(None)) else pytz.timezone("UTC")
    )
    for format in DATE_FORMATS:
        try:
            dt_object = timezone.localize(datetime.strptime(dt_string, format))
            break
        except ValueError:
            pass
        except TypeError:
            # Something wasn't right with the provided string, just return it as it was
            dt_object = dt_string

    return dt_object


def convert_to_time(
    device: str,
    data: Any,
    tz: str = "UTC",
    expression: str = None,
    parent: str = None,
    subkey: str = None,
    callback: Any = None,
) -> None:
    """Find and convert all strings resembling timestamps."""
    expression = (
        expression or r"\d{2,4}[-\/]\d{1,2}[-\/]\d{1,4} \d{1,2}:\d{1,2}:\d{1,2}"
    )
    if hasattr(data, "__dict__"):
        data = data.__dict__

    if isinstance(subkey, type(None)):
        parent = subkey
    else:
        if isinstance(parent, type(None)):
            parent = subkey
        else:
            parent += f";;{subkey}"

    for key in data:
        if key.startswith("_") or key == "devices":
            continue

        if not key in data:
            continue

        hits = []
        value = data[key]

        if isinstance(value, tuple(all_utils)) or isinstance(value, dict):
            convert_to_time(
                device=device,
                data=value,
                tz=tz,
                expression=expression,
                parent=parent,
                subkey=key,
                callback=callback,
            )
        elif isinstance(value, str):
            hits = re.findall(expression, value)
        else:
            continue

        if len(hits) == 1:
            newtime = string_to_time(hits[0], tz)
            callback(device, parent, key, newtime)



"""Blade information."""
from __future__ import annotations

from typing import Any

from .landroid_class import LDict


class Blades(LDict):
    """Blade information."""

    def __init__(
        self,
        data: Any | None = None,
    ) -> None:
        """Initialize blade object."""
        super().__init__()
        from ..helpers import string_to_time

        if isinstance(data, type(None)):
            return

        if "blade_work_time" in data:
            # Total time with blades on in minutes
            self["total_on"] = (
                int(data["blade_work_time"])
                if not isinstance(data["blade_work_time_reset"], type(None))
                else None
            )
        else:
            self["total_on"] = None

        if "blade_work_time_reset" in data:
            # Blade time reset at minutes
            self["reset_at"] = (
                int(data["blade_work_time_reset"])
                if not isinstance(data["blade_work_time_reset"], type(None))
                else None
            )
        else:
            self["reset_at"] = None

        if "blade_work_time_reset_at" in data:
            # Blade time reset time and date
            self["reset_time"] = (
                string_to_time(data["blade_work_time_reset_at"], data["time_zone"])
                if not isinstance(data["blade_work_time_reset_at"], type(None))
                else None
            )
        else:
            self["reset_time"] = None

        # Calculate blade data since reset, if possible
        if self["reset_at"] and self["total_on"]:
            # Blade time since last reset
            self["current_on"] = int(self["total_on"] - self["reset_at"])
        else:
            self["current_on"] = self["total_on"]



"""Device orientation."""
from __future__ import annotations

from .landroid_class import LDict


class Orientation(LDict):
    """Device orientation class."""

    def __init__(self, data: list) -> None:
        """Initialize orientation object."""
        super().__init__()
        if not data:
            return

        self["pitch"] = data[0]
        self["roll"] = data[1]
        self["yaw"] = data[2]



# -*- coding: utf-8 -*-

import sys

import setuptools

if sys.version_info < (3, 9, 0):
    sys.exit("The pyWorxcloud module requires Python 3.10.0 or later")

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyworxcloud",
    description="Landroid cloud API library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Morten Trab",
    author_email="morten@trab.dk",
    license="MIT",
    url="https://github.com/mtrab/pyworxcloud",
    packages=setuptools.find_packages(),
    project_urls={
        "Bug Tracker": "https://github.com/mtrab/pyworxcloud/issues",
    },
    install_requires=[
        "paho-mqtt>=1.6.1",
        "pyOpenSSL>=22.0.0",
        "ratelimit>=2.2.1",
        "requests>=2.26.0",
        "pytz>=2022.1",
    ],
)



"""Landroid Cloud exception definitions."""
from __future__ import annotations


class NoPartymodeError(Exception):
    """Define an error when partymode is not supported."""


class NoOneTimeScheduleError(Exception):
    """Define an error when OTS is not supported."""


class OfflineError(Exception):
    """Define an offline error."""


class TokenError(Exception):
    """Define an token error."""


class APIException(Exception):
    """Define an error when communicating with the API."""


class TimeoutException(Exception):
    """Define a timeout error."""


class RequestException(Exception):
    """Define a request exception."""


class MQTTException(Exception):
    """Define a MQTT exception."""


class RateLimit(Exception):
    """Defines a ratelimit exception."""

    def __init__(self, message):
        """Custom ratelimit exception class"""
        super(RateLimit, self).__init__(message)
        self.message = message


# Exception classes for URL requests
class RequestError(Exception):
    """Define a bad request error (400)."""


class AuthorizationError(Exception):
    """Represents an authorization error (401)."""


class ForbiddenError(Exception):
    """Represents an access forbidden error (403)."""


class NotFoundError(Exception):
    """Represents a not found error (404)."""


class TooManyRequestsError(Exception):
    """Represents a error when request quota have been exceeded (429)."""


class InternalServerError(Exception):
    """Represents an internal server error (500)."""


class ServiceUnavailableError(Exception):
    """Represents a service unavailable error (503)."""



"""Warranty information handler."""
from __future__ import annotations

from datetime import datetime
from typing import Any

import pytz

from .landroid_class import LDict


class Warranty(LDict):
    """Class for handling warranty information."""

    def __init__(self, data: Any) -> None:
        from ..helpers.time_format import string_to_time

        super().__init__()

        self["expires_at"] = string_to_time(
            data["warranty_expires_at"], data["time_zone"]
        )
        self["registered"] = data["warranty_registered"]
        self["expired"] = (
            bool(
                self["expires_at"]
                < datetime.now().astimezone(pytz.timezone(data["time_zone"]))
            )
            if not isinstance(self["expires_at"], type(None))
            else None
        )



"""Handling lawn parameters."""
from __future__ import annotations

from .landroid_class import LDict


class Lawn(LDict):
    """Handler for lawn parameters."""

    def __init__(self, perimeter, size) -> None:
        super().__init__()

        self["perimeter"] = perimeter
        self["size"] = size



"""Location information."""
from __future__ import annotations

from .landroid_class import LDict


class Location(LDict):
    """GPS location."""

    def __init__(self, latitude: float = None, longitude: float = None):
        """Initialize location object."""
        super().__init__()

        if not latitude or not longitude:
            return

        self["latitude"] = latitude
        self["longitude"] = longitude



"""Landroid Cloud callback events"""
# pylint: disable=unnecessary-lambda
from __future__ import annotations

import logging
from enum import IntEnum
from typing import Any


class LandroidEvent(IntEnum):
    """Enum for Landroid event types."""

    DATA_RECEIVED = 0
    MQTT_CONNECTION = 1
    MQTT_RATELIMIT = 2
    MQTT_PUBLISH = 3
    LOG = 4


_LOGGER = logging.getLogger("pyworxcloud.events")


def check_syntax(args: dict[str, Any], objs: list[str], expected_type: Any) -> bool:
    """Check if the object is of the expected type."""
    _LOGGER.debug("Checking %s against %s", objs, args)
    for obj in objs:
        if not obj in args:
            _LOGGER.debug("%s was not found in %s", obj, args)
            return False
        if not isinstance(args[obj], expected_type):
            _LOGGER.debug(
                "%s was of type %s and not as expected", obj, type(obj), expected_type
            )
            return False

    return True


class EventHandler:
    """Event handler for Landroid Cloud."""

    __events: dict[LandroidEvent, Any] = {}

    def __init__(self) -> None:
        """Initialize the event handler object."""

    def set_handler(self, event: LandroidEvent, func: Any) -> None:
        """Set handler for a LandroidEvent"""
        self.__events.update({event: func})

    def del_handler(self, event: LandroidEvent) -> None:
        """Remove a handler for a LandroidEvent."""
        self.__events.pop(event)

    def call(self, event: LandroidEvent, **kwargs) -> bool:
        """Call a handler if it was set."""
        if not event in self.__events:
            # Event was not set
            return False

        if LandroidEvent.DATA_RECEIVED == event:
            from .utils.devices import DeviceHandler

            if not check_syntax(kwargs, ["name"], str) or not check_syntax(
                kwargs, ["device"], DeviceHandler
            ):
                _LOGGER.warning(
                    "requirements for attributes was not fulfilled, not sending event!"
                )
                return False

            self.__events[event](name=kwargs["name"], device=kwargs["device"])
            return True
        elif LandroidEvent.MQTT_CONNECTION == event:
            if not check_syntax(kwargs, ["state"], bool):
                return False

            self.__events[event](state=kwargs["state"])
            return True
        elif LandroidEvent.MQTT_RATELIMIT == event:
            if not check_syntax(kwargs, ["message"], str):
                return False

            self.__events[event](message=kwargs["message"])
            return True
        elif LandroidEvent.MQTT_PUBLISH == event:
            if not check_syntax(kwargs, ["message", "device", "topic"], str):
                return False

            if not check_syntax(kwargs, ["qos"], int):
                return False

            if not check_syntax(kwargs, ["retain"], bool):
                return False

            self.__events[event](
                message=kwargs["message"],
                qos=kwargs["qos"],
                retain=kwargs["retain"],
                device=kwargs["device"],
                topic=kwargs["topic"],
            )
            return True
        elif LandroidEvent.LOG == event:
            if not check_syntax(kwargs, ["message", "level"], str):
                return False

            self.__events[event](
                message=kwargs["message"],
                level=kwargs["level"],
            )
            return True
        else:
            # Not a valid LandroidEvent
            return False



"""Supported cloud endpoints."""
from __future__ import annotations

from enum import Enum
from typing import Dict

WORX = "worx"
KRESS = "kress"
LANDXCAPE = "landxcape"
FERREX = "ferrex"


class CloudType(object):
    """Supported cloud types.

    CloudType.WORX: For Worx Landroid devices.

    CloudType.KRESS: For Kress devices.

    CloudType.LANDXCAPE: For Landxcape devices.

    CloudType.FERREX: For Aldi Ferrex devices.
    """

    class WORX(str):
        """Settings for Worx devices."""

        URL: str = "api.worxlandroid.com"

        KEY: str = "725f542f5d2c4b6a5145722a2a6a5b736e764f6e725b462e4568764d4b58755f6a767b2b76526457"

    class KRESS(str):
        """Settings for Kress devices."""

        URL: str = "api.kress-robotik.com"

        KEY: str = "5a1c6f60645658795b78416f747d7a591a494a5c6a1c4d571d194a6b595f5a7f7d7b5656771e1c5f"

    class LANDXCAPE(str):
        """Settings for Landxcape devices."""

        URL: str = "api.landxcape-services.com"

        KEY: str = "071916003330192318141c080b10131a056115181634352016310817031c0b25391c1a176a0a0102"

    class FERREX(str):
        """Settings for Aldi Ferrex devices."""

        URL: str = "api.watermelon.smartmower.cloud"

        KEY: str = "xZY9IAxGAqe1wpMRKA39M9gRkLfX6IW5zbgwCi0E"
